@@ -21,11 +21,16 @@
  */
 
 
-#ifndef GEANY_HIGHLIGHTINGMAPPINGS_H
-#define GEANY_HIGHLIGHTINGMAPPINGS_H 1
+#ifndef GEANY_HIGHLIGHTING_MAPPINGS_H
+#define GEANY_HIGHLIGHTING_MAPPINGS_H 1
 
 #include "Scintilla.h"
+#include "SciLexer.h"
 
+#include <glib.h>
+
+
+G_BEGIN_DECLS
 
 /* contains all filtypes informations in the form of:
  *  - highlighting_lexer_LANG:		the SCI lexer
@@ -168,6 +173,10 @@ static const HLStyle highlighting_styles_BASIC[] =
 {
 	{ SCE_B_DEFAULT,		"default",			FALSE },
 	{ SCE_B_COMMENT,		"comment",			FALSE },
+	{ SCE_B_COMMENTBLOCK,	"commentblock",		FALSE },
+	{ SCE_B_DOCLINE,		"docline",			FALSE },
+	{ SCE_B_DOCBLOCK,		"docblock",			FALSE },
+	{ SCE_B_DOCKEYWORD,		"dockeyword",		FALSE },
 	{ SCE_B_NUMBER,			"number",			FALSE },
 	{ SCE_B_KEYWORD,		"word",				FALSE },
 	{ SCE_B_STRING,			"string",			FALSE },
@@ -196,6 +205,27 @@ static const HLKeyword highlighting_keywords_BASIC[] =
 #define highlighting_properties_BASIC	EMPTY_PROPERTIES
 
 
+/* BATCH */
+#define highlighting_lexer_BATCH		SCLEX_BATCH
+static const HLStyle highlighting_styles_BATCH[] =
+{
+	{ SCE_BAT_DEFAULT,		"default",		FALSE },
+	{ SCE_BAT_COMMENT,		"comment",		FALSE },
+	{ SCE_BAT_LABEL,		"label",		FALSE },
+	{ SCE_BAT_WORD,			"word",			FALSE },
+	{ SCE_BAT_HIDE,			"hide",			FALSE },
+	{ SCE_BAT_COMMAND,		"command",		FALSE },
+	{ SCE_BAT_IDENTIFIER,	"identifier",	FALSE },
+	{ SCE_BAT_OPERATOR,		"operator",		FALSE }
+};
+static const HLKeyword highlighting_keywords_BATCH[] =
+{
+	{ 0, "keywords",			FALSE },
+	{ 1, "keywords_optional",	FALSE }
+};
+#define highlighting_properties_BATCH	EMPTY_PROPERTIES
+
+
 /* C */
 /* Also used by some other SCLEX_CPP-based filetypes */
 #define highlighting_lexer_C		SCLEX_CPP
@@ -206,12 +236,14 @@ static const HLStyle highlighting_styles_C[] =
 	{ SCE_C_COMMENTLINE,			"commentline",				FALSE },
 	{ SCE_C_COMMENTDOC,				"commentdoc",				FALSE },
 	{ SCE_C_PREPROCESSORCOMMENT,	"preprocessorcomment",		FALSE },
+	{ SCE_C_PREPROCESSORCOMMENTDOC,	"preprocessorcommentdoc",	FALSE },
 	{ SCE_C_NUMBER,					"number",					FALSE },
 	{ SCE_C_WORD,					"word",						FALSE },
 	{ SCE_C_WORD2,					"word2",					FALSE },
 	{ SCE_C_STRING,					"string",					FALSE },
 	{ SCE_C_STRINGRAW,				"stringraw",				FALSE },
 	{ SCE_C_CHARACTER,				"character",				FALSE },
+	{ SCE_C_USERLITERAL,			"userliteral",				FALSE },
 	{ SCE_C_UUID,					"uuid",						FALSE },
 	{ SCE_C_PREPROCESSOR,			"preprocessor",				FALSE },
 	{ SCE_C_OPERATOR,				"operator",					FALSE },
@@ -226,7 +258,9 @@ static const HLStyle highlighting_styles_C[] =
 	{ SCE_C_COMMENTDOCKEYWORD,		"commentdockeyword",		FALSE },
 	{ SCE_C_COMMENTDOCKEYWORDERROR,	"commentdockeyworderror",	FALSE },
 	/* used for local structs and typedefs */
-	{ SCE_C_GLOBALCLASS,			"globalclass",				FALSE }
+	{ SCE_C_GLOBALCLASS,			"globalclass",				FALSE },
+	{ SCE_C_TASKMARKER,				"taskmarker",				FALSE },
+	{ SCE_C_ESCAPESEQUENCE,			"escapesequence",			FALSE }
 };
 static const HLKeyword highlighting_keywords_C[] =
 {
@@ -234,7 +268,7 @@ static const HLKeyword highlighting_keywords_C[] =
 	/* SCI_SETKEYWORDS = 1 - secondary + global tags file types, see below */
 	{ 1, "secondary",	TRUE },
 	{ 2, "docComment",	FALSE }
-	/* SCI_SETKEYWORDS = 3 is for current session types - see editor_lexer_get_type_keyword_idx() */
+	/* SCI_SETKEYWORDS = 3 is for current session types - see document_highlight_tags() */
 };
 static const HLProperty highlighting_properties_C[] =
 {
@@ -299,6 +333,34 @@ static const HLKeyword highlighting_keywords_CMAKE[] =
 };
 #define highlighting_properties_CMAKE	EMPTY_PROPERTIES
 
+/* CoffeeScript */
+#define highlighting_lexer_COFFEESCRIPT		SCLEX_COFFEESCRIPT
+static const HLStyle highlighting_styles_COFFEESCRIPT[] =
+{
+	{ SCE_COFFEESCRIPT_DEFAULT,					"default",				FALSE },
+	{ SCE_COFFEESCRIPT_COMMENTLINE,				"commentline",			FALSE },
+	{ SCE_COFFEESCRIPT_NUMBER,					"number",				FALSE },
+	{ SCE_COFFEESCRIPT_WORD,					"word",					FALSE },
+	{ SCE_COFFEESCRIPT_STRING,					"string",				FALSE },
+	{ SCE_COFFEESCRIPT_CHARACTER,				"character",			FALSE },
+	{ SCE_COFFEESCRIPT_OPERATOR,				"operator",				FALSE },
+	{ SCE_COFFEESCRIPT_IDENTIFIER,				"identifier",			FALSE },
+	{ SCE_COFFEESCRIPT_STRINGEOL,				"stringeol",			FALSE },
+	{ SCE_COFFEESCRIPT_REGEX,					"regex",				FALSE },
+	{ SCE_COFFEESCRIPT_WORD2,					"word2",				FALSE },
+	{ SCE_COFFEESCRIPT_GLOBALCLASS,				"globalclass",			FALSE },
+	{ SCE_COFFEESCRIPT_COMMENTBLOCK,			"commentblock",			FALSE },
+	{ SCE_COFFEESCRIPT_VERBOSE_REGEX,			"verbose_regex",		FALSE },
+	{ SCE_COFFEESCRIPT_VERBOSE_REGEX_COMMENT,	"verbose_regex_comment",FALSE }
+};
+static const HLKeyword highlighting_keywords_COFFEESCRIPT[] =
+{
+	{ 0, "primary",		FALSE },
+	{ 1, "secondary",	FALSE },
+	{ 3, "globalclass",	FALSE }
+};
+#define highlighting_properties_COFFEESCRIPT	EMPTY_PROPERTIES
+
 
 /* CSS */
 #define highlighting_lexer_CSS			SCLEX_CSS
@@ -405,7 +467,7 @@ static const HLKeyword highlighting_keywords_D[] =
 	/* SCI_SETKEYWORDS = 1 - secondary + global tags file types */
 	{ 1, "secondary",	TRUE },
 	{ 2, "docComment",	FALSE },
-	/* SCI_SETKEYWORDS = 3 is for current session types - see editor_lexer_get_type_keyword_idx() */
+	/* SCI_SETKEYWORDS = 3 is for current session types - see document_highlight_tags() */
 	{ 4, "types",		FALSE },
 };
 #define highlighting_properties_D		EMPTY_PROPERTIES
@@ -590,31 +652,46 @@ static const HLKeyword highlighting_keywords_FORTH[] =
 #define highlighting_properties_FORTRAN		highlighting_properties_F77
 
 
+/* Go */
+#define highlighting_lexer_GO		SCLEX_CPP
+#define highlighting_styles_GO		highlighting_styles_C
+#define highlighting_keywords_GO	highlighting_keywords_C
+#define highlighting_properties_GO	highlighting_properties_C
+
+
 /* Haskell */
 #define highlighting_lexer_HASKELL			SCLEX_HASKELL
 static const HLStyle highlighting_styles_HASKELL[] =
 {
-	{ SCE_HA_DEFAULT,		"default",			FALSE },
-	{ SCE_HA_COMMENTLINE,	"commentline",		FALSE },
-	{ SCE_HA_COMMENTBLOCK,	"commentblock",		FALSE },
-	{ SCE_HA_COMMENTBLOCK2,	"commentblock2",	FALSE },
-	{ SCE_HA_COMMENTBLOCK3,	"commentblock3",	FALSE },
-	{ SCE_HA_NUMBER,		"number",			FALSE },
-	{ SCE_HA_KEYWORD,		"keyword",			FALSE },
-	{ SCE_HA_IMPORT,		"import",			FALSE },
-	{ SCE_HA_STRING,		"string",			FALSE },
-	{ SCE_HA_CHARACTER,		"character",		FALSE },
-	{ SCE_HA_CLASS,			"class",			FALSE },
-	{ SCE_HA_OPERATOR,		"operator",			FALSE },
-	{ SCE_HA_IDENTIFIER,	"identifier",		FALSE },
-	{ SCE_HA_INSTANCE,		"instance",			FALSE },
-	{ SCE_HA_CAPITAL,		"capital",			FALSE },
-	{ SCE_HA_MODULE,		"module",			FALSE },
-	{ SCE_HA_DATA,			"data",				FALSE }
+	{ SCE_HA_DEFAULT,				"default",				FALSE },
+	{ SCE_HA_COMMENTLINE,			"commentline",			FALSE },
+	{ SCE_HA_COMMENTBLOCK,			"commentblock",			FALSE },
+	{ SCE_HA_COMMENTBLOCK2,			"commentblock2",		FALSE },
+	{ SCE_HA_COMMENTBLOCK3,			"commentblock3",		FALSE },
+	{ SCE_HA_NUMBER,				"number",				FALSE },
+	{ SCE_HA_KEYWORD,				"keyword",				FALSE },
+	{ SCE_HA_IMPORT,				"import",				FALSE },
+	{ SCE_HA_STRING,				"string",				FALSE },
+	{ SCE_HA_CHARACTER,				"character",			FALSE },
+	{ SCE_HA_CLASS,					"class",				FALSE },
+	{ SCE_HA_OPERATOR,				"operator",				FALSE },
+	{ SCE_HA_IDENTIFIER,			"identifier",			FALSE },
+	{ SCE_HA_INSTANCE,				"instance",				FALSE },
+	{ SCE_HA_CAPITAL,				"capital",				FALSE },
+	{ SCE_HA_MODULE,				"module",				FALSE },
+	{ SCE_HA_DATA,					"data",					FALSE },
+	{ SCE_HA_PRAGMA,				"pragma",				FALSE },
+	{ SCE_HA_PREPROCESSOR,			"preprocessor",			FALSE },
+	{ SCE_HA_STRINGEOL,				"stringeol",			FALSE },
+	{ SCE_HA_RESERVED_OPERATOR,		"reserved_operator",	FALSE },
+	{ SCE_HA_LITERATE_COMMENT,		"literate_comment",		FALSE },
+	{ SCE_HA_LITERATE_CODEDELIM,	"literate_codedelim",	FALSE }
 };
 static const HLKeyword highlighting_keywords_HASKELL[] =
 {
-	{ 0, "keywords",	FALSE }
+	{ 0, "keywords",		   FALSE },
+	{ 1, "ffi",				   FALSE },
+	{ 2, "reserved_operators", FALSE }
 };
 #define highlighting_properties_HASKELL		EMPTY_PROPERTIES
 
@@ -787,7 +864,7 @@ static const HLKeyword highlighting_keywords_JAVA[] =
 	/* SCI_SETKEYWORDS = 1 - secondary + global tags file types, see below */
 	{ 1, "secondary",	TRUE },
 	{ 2, "doccomment",	FALSE },
-	/* SCI_SETKEYWORDS = 3 is for current session types - see editor_lexer_get_type_keyword_idx() */
+	/* SCI_SETKEYWORDS = 3 is for current session types - see document_highlight_tags() */
 	{ 4, "typedefs",	FALSE }
 };
 #define highlighting_properties_JAVA	highlighting_properties_C
@@ -1005,7 +1082,7 @@ static const HLKeyword highlighting_keywords_OBJECTIVEC[] =
 	/* SCI_SETKEYWORDS = 1 - secondary + global tags file types, see below */
 	{ 1, "secondary",	TRUE },
 	{ 2, "docComment",	FALSE }
-	/* SCI_SETKEYWORDS = 3 is for current session types - see editor_lexer_get_type_keyword_idx() */
+	/* SCI_SETKEYWORDS = 3 is for current session types - see document_highlight_tags() */
 };
 #define highlighting_properties_OBJECTIVEC	highlighting_properties_C
 
@@ -1126,6 +1203,40 @@ static const HLStyle highlighting_styles_PO[] =
 #define highlighting_properties_PO	EMPTY_PROPERTIES
 
 
+/* PowerShell */
+#define highlighting_lexer_POWERSHELL		SCLEX_POWERSHELL
+static const HLStyle highlighting_styles_POWERSHELL[] =
+{
+	{ SCE_POWERSHELL_DEFAULT,			"default",				FALSE },
+	{ SCE_POWERSHELL_COMMENT,			"comment",				FALSE },
+	{ SCE_POWERSHELL_STRING,			"string",				FALSE },
+	{ SCE_POWERSHELL_CHARACTER,			"character",			FALSE },
+	{ SCE_POWERSHELL_NUMBER,			"number",				FALSE },
+	{ SCE_POWERSHELL_VARIABLE,			"variable",				FALSE },
+	{ SCE_POWERSHELL_OPERATOR,			"operator",				FALSE },
+	{ SCE_POWERSHELL_IDENTIFIER,		"identifier",			FALSE },
+	{ SCE_POWERSHELL_KEYWORD,			"keyword",				FALSE },
+	{ SCE_POWERSHELL_CMDLET,			"cmdlet",				FALSE },
+	{ SCE_POWERSHELL_ALIAS,				"alias",				FALSE },
+	{ SCE_POWERSHELL_FUNCTION,			"function",				FALSE },
+	{ SCE_POWERSHELL_USER1,				"user1",				FALSE },
+	{ SCE_POWERSHELL_COMMENTSTREAM,		"commentstream",		FALSE },
+	{ SCE_POWERSHELL_HERE_STRING,		"here_string",			FALSE },
+	{ SCE_POWERSHELL_HERE_CHARACTER,	"here_character",		FALSE },
+	{ SCE_POWERSHELL_COMMENTDOCKEYWORD,	"commentdockeyword",	FALSE },
+};
+static const HLKeyword highlighting_keywords_POWERSHELL[] =
+{
+	{ 0, "keywords",	FALSE },
+	{ 1, "cmdlets",		FALSE },
+	{ 2, "aliases",		FALSE },
+	{ 3, "functions",	FALSE },
+	{ 4, "user1",		FALSE },
+	{ 5, "docComment",	FALSE },
+};
+#define highlighting_properties_POWERSHELL	EMPTY_PROPERTIES
+
+
 /* Python */
 #define highlighting_lexer_PYTHON		SCLEX_PYTHON
 static const HLStyle highlighting_styles_PYTHON[] =
@@ -1226,6 +1337,44 @@ static const HLKeyword highlighting_keywords_RUBY[] =
 };
 #define highlighting_properties_RUBY	EMPTY_PROPERTIES
 
+/* Rust */
+#define highlighting_lexer_RUST		SCLEX_RUST
+static const HLStyle highlighting_styles_RUST[] =
+{
+	{ SCE_RUST_DEFAULT,				"default",					FALSE },
+	{ SCE_RUST_COMMENTBLOCK,		"commentblock",				FALSE },
+	{ SCE_RUST_COMMENTLINE,			"commentline",				FALSE },
+	{ SCE_RUST_COMMENTBLOCKDOC,		"commentblockdoc",			FALSE },
+	{ SCE_RUST_COMMENTLINEDOC,		"commentlinedoc",			FALSE },
+	{ SCE_RUST_NUMBER,				"number",					FALSE },
+	{ SCE_RUST_WORD,				"word",						FALSE },
+	{ SCE_RUST_WORD2,				"word2",					FALSE },
+	{ SCE_RUST_WORD3,				"word3",					FALSE },
+	{ SCE_RUST_WORD4,				"word4",					FALSE },
+	{ SCE_RUST_WORD5,				"word5",					FALSE },
+	{ SCE_RUST_WORD6,				"word6",					FALSE },
+	{ SCE_RUST_WORD7,				"word7",					FALSE },
+	{ SCE_RUST_STRING,				"string",					FALSE },
+	{ SCE_RUST_STRINGR,				"stringraw",				FALSE },
+	{ SCE_RUST_CHARACTER,			"character",				FALSE },
+	{ SCE_RUST_OPERATOR,			"operator",					FALSE },
+	{ SCE_RUST_IDENTIFIER,			"identifier",				FALSE },
+	{ SCE_RUST_LIFETIME,			"lifetime",					FALSE },
+	{ SCE_RUST_MACRO,				"macro",					FALSE },
+	{ SCE_RUST_LEXERROR,			"lexerror",					FALSE },
+	{ SCE_RUST_BYTESTRING,			"bytestring",				FALSE },
+	{ SCE_RUST_BYTESTRINGR,			"bytestringr",				FALSE },
+	{ SCE_RUST_BYTECHARACTER,		"bytecharacter",			FALSE }
+};
+static const HLKeyword highlighting_keywords_RUST[] =
+{
+	{ 0, "primary",		FALSE },
+	/* SCI_SETKEYWORDS = 1 - secondary + global tags file types */
+	{ 1, "secondary",	TRUE },
+	{ 2, "tertiary",	FALSE },
+	/* SCI_SETKEYWORDS = 3 is for current session types - see document_highlight_tags() */
+};
+#define highlighting_properties_RUST		EMPTY_PROPERTIES
 
 /* SH */
 #define highlighting_lexer_SH		SCLEX_BASH
@@ -1274,7 +1423,8 @@ static const HLStyle highlighting_styles_SQL[] =
 	{ SCE_SQL_SQLPLUS,					"sqlplus",					FALSE },
 	{ SCE_SQL_SQLPLUS_PROMPT,			"sqlplus_prompt",			FALSE },
 	{ SCE_SQL_SQLPLUS_COMMENT,			"sqlplus_comment",			FALSE },
-	{ SCE_SQL_QUOTEDIDENTIFIER,			"quotedidentifier",			FALSE }
+	{ SCE_SQL_QUOTEDIDENTIFIER,			"quotedidentifier",			FALSE },
+	{ SCE_SQL_QOPERATOR,				"qoperator",				FALSE }
 	/* these are for user-defined keywords we don't set yet */
 	/*{ SCE_SQL_USER1,					"user1",					FALSE },
 	{ SCE_SQL_USER2,					"user2",					FALSE },
@@ -1369,6 +1519,7 @@ static const HLStyle highlighting_styles_VHDL[] =
 	{ SCE_VHDL_DEFAULT,			"default",				FALSE },
 	{ SCE_VHDL_COMMENT,			"comment",				FALSE },
 	{ SCE_VHDL_COMMENTLINEBANG,	"comment_line_bang",	FALSE },
+	{ SCE_VHDL_BLOCK_COMMENT,	"block_comment",		FALSE },
 	{ SCE_VHDL_NUMBER,			"number",				FALSE },
 	{ SCE_VHDL_STRING,			"string",				FALSE },
 	{ SCE_VHDL_OPERATOR,		"operator",				FALSE },
@@ -1412,13 +1563,19 @@ static const HLStyle highlighting_styles_VERILOG[] =
 	{ SCE_V_OPERATOR,			"operator",				FALSE },
 	{ SCE_V_IDENTIFIER,			"identifier",			FALSE },
 	{ SCE_V_STRINGEOL,			"stringeol",			FALSE },
-	{ SCE_V_USER,				"userword",				FALSE }
+	{ SCE_V_USER,				"userword",				FALSE },
+	{ SCE_V_COMMENT_WORD,		"comment_word",			FALSE },
+	{ SCE_V_INPUT,				"input",				FALSE },
+	{ SCE_V_OUTPUT,				"output",				FALSE },
+	{ SCE_V_INOUT,				"inout",				FALSE },
+	{ SCE_V_PORT_CONNECT,		"port_connect",			FALSE }
 };
 static const HLKeyword highlighting_keywords_VERILOG[] =
 {
-	{ 0, "word",	FALSE },
-	{ 1, "word2",	FALSE },
-	{ 2, "word3",	FALSE }
+	{ 0, "word",		FALSE },
+	{ 1, "word2",		FALSE },
+	{ 2, "word3",		FALSE },
+	{ 4, "docComment",	FALSE }
 };
 #define highlighting_properties_VERILOG		EMPTY_PROPERTIES
 
@@ -1455,4 +1612,12 @@ static const HLKeyword highlighting_keywords_YAML[] =
 #define highlighting_properties_YAML	EMPTY_PROPERTIES
 
 
-#endif /* guard */
+/* Zephir */
+#define highlighting_lexer_ZEPHIR		SCLEX_PHPSCRIPT
+#define highlighting_styles_ZEPHIR		highlighting_styles_PHP
+#define highlighting_keywords_ZEPHIR	highlighting_keywords_PHP
+#define highlighting_properties_ZEPHIR	highlighting_properties_PHP
+
+G_END_DECLS
+
+#endif /* GEANY_HIGHLIGHTING_MAPPINGS_H */
