@@ -76,14 +76,33 @@
  * should be efficient enough.
  */
 
-
-#include "geany.h"		/* necessary for utils.h, otherwise use gtk/gtk.h */
-#include <stdlib.h>		/* only for atoi() */
-#include "support.h"	/* only for _("text") */
-#include "utils.h"		/* only for foreach_*, utils_get_setting_*(). Stash should not depend on Geany. */
-
 #include "stash.h"
 
+#include "support.h" /* only for _("text") */
+#include "utils.h"   /* only for foreach_*, utils_get_setting_*(). Stash should not depend on Geany. */
+
+#include <stdlib.h> /* only for atoi() */
+
+
+/* GTK3 removed ComboBoxEntry, but we need a value to differentiate combo box with and
+ * without entries, and it must not collide with other GTypes */
+#ifdef GTK_TYPE_COMBO_BOX_ENTRY
+#	define TYPE_COMBO_BOX_ENTRY GTK_TYPE_COMBO_BOX_ENTRY
+#else /* !GTK_TYPE_COMBO_BOX_ENTRY */
+#	define TYPE_COMBO_BOX_ENTRY get_combo_box_entry_type()
+static GType get_combo_box_entry_type(void)
+{
+	static volatile gsize type = 0;
+	if (g_once_init_enter(&type))
+	{
+		GType g_type = g_type_register_static_simple(GTK_TYPE_COMBO_BOX, "dummy-combo-box-entry",
+				sizeof(GtkComboBoxClass), NULL, sizeof(GtkComboBox), NULL, G_TYPE_FLAG_ABSTRACT);
+		g_once_init_leave(&type, g_type);
+	}
+	return type;
+}
+#endif /* !GTK_TYPE_COMBO_BOX_ENTRY */
+
 
 struct StashPref
 {
@@ -255,6 +274,7 @@ static void keyfile_action(SettingAction action, StashGroup *group, GKeyFile *ke
  * so that all Stash settings are initialized to defaults.
  * @param group .
  * @param keyfile Usually loaded from a configuration file first. */
+GEANY_API_SYMBOL
 void stash_group_load_from_key_file(StashGroup *group, GKeyFile *keyfile)
 {
 	keyfile_action(SETTING_READ, group, keyfile);
@@ -265,6 +285,7 @@ void stash_group_load_from_key_file(StashGroup *group, GKeyFile *keyfile)
  * @a keyfile is usually written to a configuration file afterwards.
  * @param group .
  * @param keyfile . */
+GEANY_API_SYMBOL
 void stash_group_save_to_key_file(StashGroup *group, GKeyFile *keyfile)
 {
 	keyfile_action(SETTING_WRITE, group, keyfile);
@@ -279,6 +300,7 @@ void stash_group_save_to_key_file(StashGroup *group, GKeyFile *keyfile)
  * @return @c TRUE if a key file could be loaded.
  * @see stash_group_load_from_key_file().
  **/
+GEANY_API_SYMBOL
 gboolean stash_group_load_from_file(StashGroup *group, const gchar *filename)
 {
 	GKeyFile *keyfile;
@@ -303,6 +325,7 @@ gboolean stash_group_load_from_file(StashGroup *group, const gchar *filename)
  *         failed operation is returned.
  * @see stash_group_save_to_key_file().
  **/
+GEANY_API_SYMBOL
 gint stash_group_save_to_file(StashGroup *group, const gchar *filename,
 		GKeyFileFlags flags)
 {
@@ -327,6 +350,7 @@ gint stash_group_save_to_file(StashGroup *group, const gchar *filename,
 /** Creates a new group.
  * @param name Name used for @c GKeyFile group.
  * @return Group. */
+GEANY_API_SYMBOL
 StashGroup *stash_group_new(const gchar *name)
 {
 	StashGroup *group = g_new0(StashGroup, 1);
@@ -342,6 +366,7 @@ StashGroup *stash_group_new(const gchar *name)
  * Useful e.g. to avoid freeing strings individually.
  * @note This is *not* called by stash_group_free().
  * @param group . */
+GEANY_API_SYMBOL
 void stash_group_free_settings(StashGroup *group)
 {
 	StashPref *entry;
@@ -363,6 +388,7 @@ void stash_group_free_settings(StashGroup *group)
 
 /** Frees a group.
  * @param group . */
+GEANY_API_SYMBOL
 void stash_group_free(StashGroup *group)
 {
 	StashPref *entry;
@@ -423,6 +449,7 @@ add_pref(StashGroup *group, GType type, gpointer setting,
  * @param setting Address of setting variable.
  * @param key_name Name for key in a @c GKeyFile.
  * @param default_value Value to use if the key doesn't exist when loading. */
+GEANY_API_SYMBOL
 void stash_group_add_boolean(StashGroup *group, gboolean *setting,
 		const gchar *key_name, gboolean default_value)
 {
@@ -435,6 +462,7 @@ void stash_group_add_boolean(StashGroup *group, gboolean *setting,
  * @param setting Address of setting variable.
  * @param key_name Name for key in a @c GKeyFile.
  * @param default_value Value to use if the key doesn't exist when loading. */
+GEANY_API_SYMBOL
 void stash_group_add_integer(StashGroup *group, gint *setting,
 		const gchar *key_name, gint default_value)
 {
@@ -448,6 +476,7 @@ void stash_group_add_integer(StashGroup *group, gint *setting,
  * @param setting Address of setting variable.
  * @param key_name Name for key in a @c GKeyFile.
  * @param default_value String to copy if the key doesn't exist when loading, or @c NULL. */
+GEANY_API_SYMBOL
 void stash_group_add_string(StashGroup *group, gchar **setting,
 		const gchar *key_name, const gchar *default_value)
 {
@@ -461,6 +490,7 @@ void stash_group_add_string(StashGroup *group, gchar **setting,
  * @param setting Address of setting variable.
  * @param key_name Name for key in a @c GKeyFile.
  * @param default_value Vector to copy if the key doesn't exist when loading. Usually @c NULL. */
+GEANY_API_SYMBOL
 void stash_group_add_string_vector(StashGroup *group, gchar ***setting,
 		const gchar *key_name, const gchar **default_value)
 {
@@ -648,12 +678,25 @@ static void handle_widget_property(GtkWidget *widget, StashPref *entry,
 	switch (action)
 	{
 		case PREF_DISPLAY:
-			g_object_set(object, name, entry->setting, NULL);
+			if (entry->setting_type == G_TYPE_BOOLEAN)
+				g_object_set(object, name, *(gboolean*)entry->setting, NULL);
+			else if (entry->setting_type == G_TYPE_INT)
+				g_object_set(object, name, *(gint*)entry->setting, NULL);
+			else if (entry->setting_type == G_TYPE_STRING)
+				g_object_set(object, name, *(gchararray*)entry->setting, NULL);
+			else if (entry->setting_type == G_TYPE_STRV)
+				g_object_set(object, name, *(gchararray**)entry->setting, NULL);
+			else
+			{
+				g_warning("Unhandled type %s for %s in %s()!", g_type_name(entry->setting_type),
+						entry->key_name, G_STRFUNC);
+			}
 			break;
 		case PREF_UPDATE:
 			if (entry->setting_type == G_TYPE_STRING)
-				g_free(entry->setting);
-			/* TODO: Which other types need freeing here? */
+				g_free(*(gchararray*)entry->setting);
+			else if (entry->setting_type == G_TYPE_STRV)
+				g_strfreev(*(gchararray**)entry->setting);
 
 			g_object_get(object, name, entry->setting, NULL);
 			break;
@@ -696,7 +739,7 @@ static void pref_action(PrefAction action, StashGroup *group, GtkWidget *owner)
 			handle_spin_button(widget, entry, action);
 		else if (entry->widget_type == GTK_TYPE_COMBO_BOX)
 			handle_combo_box(widget, entry, action);
-		else if (entry->widget_type == GTK_TYPE_COMBO_BOX_ENTRY)
+		else if (entry->widget_type == TYPE_COMBO_BOX_ENTRY)
 			handle_combo_box_entry(widget, entry, action);
 		else if (entry->widget_type == GTK_TYPE_ENTRY)
 			handle_entry(widget, entry, action);
@@ -715,6 +758,7 @@ static void pref_action(PrefAction action, StashGroup *group, GtkWidget *owner)
  * @param owner If non-NULL, used to lookup widgets by name, otherwise
  * widget pointers are assumed.
  * @see stash_group_update(). */
+GEANY_API_SYMBOL
 void stash_group_display(StashGroup *group, GtkWidget *owner)
 {
 	pref_action(PREF_DISPLAY, group, owner);
@@ -727,6 +771,7 @@ void stash_group_display(StashGroup *group, GtkWidget *owner)
  * @param owner If non-NULL, used to lookup widgets by name, otherwise
  * widget pointers are assumed.
  * @see stash_group_display(). */
+GEANY_API_SYMBOL
 void stash_group_update(StashGroup *group, GtkWidget *owner)
 {
 	pref_action(PREF_UPDATE, group, owner);
@@ -754,6 +799,7 @@ add_widget_pref(StashGroup *group, GType setting_type, gpointer setting,
  * @param default_value Value to use if the key doesn't exist when loading.
  * @param widget_id @c GtkWidget pointer or string to lookup widget later.
  * @see stash_group_add_radio_buttons(). */
+GEANY_API_SYMBOL
 void stash_group_add_toggle_button(StashGroup *group, gboolean *setting,
 		const gchar *key_name, gboolean default_value, StashWidgetID widget_id)
 {
@@ -776,6 +822,7 @@ void stash_group_add_toggle_button(StashGroup *group, gboolean *setting,
  * stash_group_add_radio_buttons(group, &which_one_setting, "which_one", BAR,
  * 	"radio_foo", FOO, "radio_bar", BAR, NULL);
  * @endcode */
+GEANY_API_SYMBOL
 void stash_group_add_radio_buttons(StashGroup *group, gint *setting,
 		const gchar *key_name, gint default_value,
 		StashWidgetID widget_id, gint enum_id, ...)
@@ -826,6 +873,7 @@ void stash_group_add_radio_buttons(StashGroup *group, gint *setting,
  * @param key_name Name for key in a @c GKeyFile.
  * @param default_value Value to use if the key doesn't exist when loading.
  * @param widget_id @c GtkWidget pointer or string to lookup widget later. */
+GEANY_API_SYMBOL
 void stash_group_add_spin_button_integer(StashGroup *group, gint *setting,
 		const gchar *key_name, gint default_value, StashWidgetID widget_id)
 {
@@ -841,6 +889,7 @@ void stash_group_add_spin_button_integer(StashGroup *group, gint *setting,
  * @param default_value Value to use if the key doesn't exist when loading.
  * @param widget_id @c GtkWidget pointer or string to lookup widget later.
  * @see stash_group_add_combo_box_entry(). */
+GEANY_API_SYMBOL
 void stash_group_add_combo_box(StashGroup *group, gint *setting,
 		const gchar *key_name, gint default_value, StashWidgetID widget_id)
 {
@@ -857,11 +906,12 @@ void stash_group_add_combo_box(StashGroup *group, gint *setting,
  * @param widget_id @c GtkWidget pointer or string to lookup widget later. */
 /* We could maybe also have something like stash_group_add_combo_box_entry_with_menu()
  * for the history list - or should that be stored as a separate setting? */
+GEANY_API_SYMBOL
 void stash_group_add_combo_box_entry(StashGroup *group, gchar **setting,
 		const gchar *key_name, const gchar *default_value, StashWidgetID widget_id)
 {
 	add_widget_pref(group, G_TYPE_STRING, setting, key_name, (gpointer)default_value,
-		GTK_TYPE_COMBO_BOX_ENTRY, widget_id);
+		TYPE_COMBO_BOX_ENTRY, widget_id);
 }
 
 
@@ -871,6 +921,7 @@ void stash_group_add_combo_box_entry(StashGroup *group, gchar **setting,
  * @param key_name Name for key in a @c GKeyFile.
  * @param default_value Value to use if the key doesn't exist when loading.
  * @param widget_id @c GtkWidget pointer or string to lookup widget later. */
+GEANY_API_SYMBOL
 void stash_group_add_entry(StashGroup *group, gchar **setting,
 		const gchar *key_name, const gchar *default_value, StashWidgetID widget_id)
 {
@@ -903,6 +954,7 @@ static GType object_get_property_type(GObject *object, const gchar *property_nam
  * @c GObject data.
  * @warning Currently only string GValue properties will be freed before setting; patch for
  * other types - see @c handle_widget_property(). */
+GEANY_API_SYMBOL
 void stash_group_add_widget_property(StashGroup *group, gpointer setting,
 		const gchar *key_name, gpointer default_value, StashWidgetID widget_id,
 		const gchar *property_name, GType type)
@@ -927,7 +979,7 @@ struct StashTreeValue
 {
 	const gchar *group_name;
 	StashPref *pref;
-	union
+	struct
 	{
 		gchararray tree_string;
 		gint tree_int;
@@ -1023,8 +1075,8 @@ static gboolean stash_tree_discard_value(GtkTreeModel *model, GtkTreePath *path,
 	StashTreeValue *value;
 
 	gtk_tree_model_get(model, iter, STASH_TREE_VALUE, &value, -1);
-	if (value->pref->setting_type == G_TYPE_STRING)
-		g_free(value->data.tree_string);
+	/* don't access value->pref as it might already have been freed */
+	g_free(value->data.tree_string);
 	g_free(value);
 
 	return FALSE;
