@@ -23,7 +23,16 @@
 #ifndef GEANY_SYMBOLS_H
 #define GEANY_SYMBOLS_H 1
 
-extern const guint TM_GLOBAL_TYPE_MASK;
+#include "document.h"
+
+#include <glib.h>
+
+G_BEGIN_DECLS
+
+const gchar *symbols_get_context_separator(gint ft_id);
+
+
+#ifdef GEANY_PRIVATE
 
 enum
 {
@@ -41,14 +50,10 @@ void symbols_reload_config_files(void);
 
 void symbols_global_tags_loaded(guint file_type_idx);
 
-GString *symbols_find_tags_as_string(GPtrArray *tags_array, guint tag_types, gint lang);
-
-const gchar *symbols_get_context_separator(gint ft_id);
+GString *symbols_find_typenames_as_string(gint lang, gboolean global);
 
 const GList *symbols_get_tag_list(GeanyDocument *doc, guint tag_types);
 
-GString *symbols_get_macro_list(gint lang);
-
 const gchar **symbols_get_html_entities(void);
 
 gboolean symbols_recreate_tag_list(GeanyDocument *doc, gint sort_mode);
@@ -63,4 +68,8 @@ gint symbols_get_current_function(GeanyDocument *doc, const gchar **tagname);
 
 gint symbols_get_current_scope(GeanyDocument *doc, const gchar **tagname);
 
-#endif
+#endif /* GEANY_PRIVATE */
+
+G_END_DECLS
+
+#endif /* GEANY_SYMBOLS_H */
